/*
 *  Smart Cache, http proxy cache server
 *  Copyright (C) 1998-2001 Radim Kolar
 *
 *    Smart Cache is Open Source Software; you may redistribute it
 *  and/or modify it under the terms of the GNU General Public
 *  License as published by the Free Software Foundation; either
 *  version 2, or (at your option) any later version.
 *
 *    This program distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 *  General Public License for more details.
 *
 *    A copy of the GNU General Public License is available as
 *  /usr/doc/copyright/GPL in the Debian GNU/Linux distribution or on
 *  the World Wide Web at http://www.gnu.org/copyleft/gpl.html. You
 *  can also obtain it by writing to the Free Software Foundation,
 *  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

import java.net.*;
import java.io.*;
import java.util.Date;
import java.util.Hashtable;
import java.util.StringTokenizer;
import java.util.Enumeration;

public class ui implements Runnable
{
    public final static String VERSION="0.4";
    public final static String ESCAPE=": %/=?&";
    
    public static final int OFF=0;
    public static final int ON=1;
    public static final int ASK=2;

    public static int uiport;
    public static InetAddress uiadr;
    public static String ui_hostname;

    public static int loader_add_missing;
    public static int loader_add_reloads;
    public static String loader_queue_file;
    public static String loader_refresh_depth;
    public static String loader_add_depth;

    public static String LOCK="-UI-LOCK-OBJECT-";
    
    private ServerSocket server;
    
ui()
{
 if(uiport==0) {
     server=null;
     return;
 }

 if(uiadr!=null)

  try{
  server=new ServerSocket(uiport,10,uiadr);
  }
  catch(IOException e) {System.err.println("[SMARTCACHE] Fatal error: Cannot bind to my UI port "+uiport+"/"+uiadr.getHostAddress()+"\n[SMARTCACHE]    Reason: "+e);System.exit(3);}
   else
  try{
    server=new ServerSocket(uiport);
  }
  catch(IOException e) {System.err.println("[SMARTCACHE] Fatal error: Cannot bind to my UI port "+uiport+"/*"+"\n[SMARTCACHE]    Reason: "+e);System.exit(3);}
  /* get localhost if not set */
  if(ui_hostname==null) ui_hostname="127.0.0.1";
}

public final void run()
 {
  ThreadGroup clients;
  uireq client;
  if(server==null) return;
  clients=new ThreadGroup("UI-clients");
  System.out.println(new Date()+" "+scache.CACHENAME+" UI "+VERSION+" ready on port "+uiport+".");
   while (true ) {
            Socket clientSocket = null;
            client=null;
            try {
                clientSocket = server.accept();
                } catch (IOException e) {
                System.err.println("[SMARTCACHE] Warning: UI Accept failed: "+e);
                continue;
            }
            if(clients.activeCount()>scache.MAXHTTPCLIENTS)
              try {
              System.err.println(new Date()+" [SMARTCACHE] Warning: UI Active clients limit ("+scache.MAXHTTPCLIENTS+") reached, request rejected.");
              clientSocket.close();}
              catch(IOException e) {}
              finally { continue;}
            client=new uireq(clientSocket);
            new Thread(clients,client).start();
   }  /* listen */
 }

public final static String process(String URL)
{
  if(URL.startsWith("/loader/setdepth?"))
  {
      String t=getparam(URL,"type");
      URL=getparam(URL,"depth");
      if(t!=null && URL!=null)
      {
	  if(t.equals("add"))
	      loader_add_depth=URL;
	  else
	      loader_refresh_depth=URL;

	  return "Default depth was changed.";
      }
      else
	  return "type and depth must be both set!";
  }
  else
  if(URL.startsWith("/loader/confirm/add?"))
  {
      URL=getparam(URL,"url");
      return "You have requested a URL: <em><a href=\""+URL+"\">"+URL+"</a></em> which is not in cache and can not be downloaded from remote server now. Do you want to add this URL to Smart Cache Loader queue for later batch retrieve?<p>\n"+
             "<form method=get action=\"/loader/add\">\n"+
	     "URL: <input type=text name=url size=40 value=\""+URL+"\"><br>\n"+
	     "Fetch depth: <input type=text name=depth maxlength=3 size=3 value=\""+loader_add_depth+"\"><br>\n"+
	     "Loader options: <input type=text size=20 name=options><br>\n"+
	     "<input type=submit value=\"Yes, Add it\">\n"+
	     "</form>\n"+
	     "<p><a href=\"/ui/add?off\">[X] Do not show this dialog again.</a>\n<p><a href=\"/ui/add?on\">[X] Next time add without asking.</a>\n"+
	     "<p><form method=get action=\"/loader/setdepth\">\n"+
	     "Set default depth for load to: <input type=text name=depth maxlength=3 size=3 value=\""+loader_add_depth+"\"><input type=submit value=OK>\n<input type=hidden name=type value=add></form>";
  } else
  if(URL.startsWith("/loader/confirm/refresh?"))
  {
      URL=getparam(URL,"url");
      return "You have requested a forced refresh of URL: <em><a href=\""+URL+"\">"+URL+"</a></em> which can not be done from remote server now. If this reforced refresh was auto requested by your browser (using Bookmarks), click on URL to continue.\n<p>Do you want to add this URL to Smart Cache Loader queue for later batch retrieve?<p>\n"+
             "<form method=get action=\"/loader/add\">\n"+
	     "URL: <input type=text name=url size=40 value=\""+URL+"\"><br>\n"+
	     "Refresh depth: <input type=text name=depth maxlength=3 size=3 value=\""+loader_refresh_depth+"\"><br>\n"+
	     "Loader options: <input type=text size=20 name=options><br>\n"+
	     "<input type=submit value=\"Yes, Add it\">\n"+
	     "</form>\n"+
	     "<p><a href=\"/ui/refresh?off\">[X] Do not show this dialog again.</a>\n<p><a href=\"/ui/refresh?on\">[X] Next time add to queue without asking.</a>\n"+
	     "<p><form method=get action=\"/loader/setdepth\">\n"+
	     "Set default depth for refresh to: <input type=text name=depth maxlength=3 size=3 value=\""+loader_refresh_depth+"\"><input type=submit value=OK>\n<input type=hidden name=type value=refresh></form>";
  } else
  if(URL.startsWith("/loader/add?") || URL.startsWith("/loader/refresh?"))
  {
	  String url=getparam(URL,"url");
	  String depth=getparam(URL,"depth");
	  String options=getparam(URL,"options");
	  if(depth==null) 
	      if(URL.startsWith("/loader/add"))
		  depth=loader_add_depth;
	      else
		  depth=loader_refresh_depth;
	  if(options==null) options="";
	  if(url==null)
	      return "Error: Can not add null URL!";
	  if(loader_queue_file==null)
	      return "Can not add URL! Error: loader_queue_file is not defined in scache.cnf.";
	  synchronized(LOCK)
	  {
	      Hashtable q=load_queue();
	      q.put(url,"depth="+depth+' '+options);
	      try
	      {
	         save_queue(q);
		 return "Requested URL was sucessfully added to Smart Cache loader queue file.<p>Run: \"java loader @loader_queue_file\" for batch fetch. Delete loader_queue_file after that.<p><a href=\"/loader/queue/show\">Show current queue</a>.";
	      }
	      catch (IOException dd)
	      {
		  return "Error writing to loader queue file.";
	      }
	  }
  } else
  if(URL.startsWith("/loader/queue/show"))
  {
     StringBuffer result=new StringBuffer(512);
     Hashtable q;
     synchronized(LOCK)
     {
       q=load_queue();
     }
     result.append("<h2>Loader Queue</h2>\n");
     for(Enumeration en=q.keys();en.hasMoreElements();)
     {
	 URL=(String)en.nextElement();
	 result.append("<a href=\"/loader/queue/delete?url=");
	 result.append(escape(URL));
	 result.append("\">Del</a> <a href=\"");
	 result.append(URL);
	 result.append("\">");
	 result.append(URL);
	 result.append("</a> (");
	 result.append(q.get(URL));
	 result.append(")<br>\n");
     }
     return result.toString();
  } else
  if(URL.startsWith("/loader/queue/delete?"))
  {
     String url=getparam(URL,"url");
     if(url==null)
	 return "Can not delete null URL from queue!";
     synchronized(LOCK)
     {
        Hashtable q=load_queue();
        q.remove(url);
	try
	{
	    save_queue(q);
	    return "Requested URL has been removed from queue (if exists there).<p><a href=\"/loader/queue/show\">Show queue</a>.";
	}
	catch (IOException e)
	{
	    return "Error updating loader queue file.";
	}
     }
  } else
  if(URL.startsWith("/store/browse?"))
  {
      boolean del=false;
      String bURL;
      bURL=getparam(URL,"showdel");
      if(bURL!=null)
	  if(bURL.length()>0)
	      del=true;
      URL=getparam(URL,"dir");

      if(URL==null)
      {
	  return "dir argument is required!";
      }
      if(URL.indexOf("..")>-1) 
	  return "Unsecure directory name.";
      StringBuffer ans=new StringBuffer(2048);
      ans.append("<h2>Browsing cache directory ");
      ans.append(URL);
      ans.append("</h2>\n");
      if(del==false)
      {
	  ans.append("<a href=\"/store/browse?dir=");
	  ans.append(escape(URL));
	  ans.append("&showdel=true\">Enable Delete in this directory</a><br> \n");
      }

      File f=new File(mgr.cache_dir+URL);
      if(URL.length()==0 || !f.isDirectory())
	  ans.append("This is not a valid directory!");
      else
      {
	  File z=new File(URL.substring(0,URL.length()-1));
	  /* dir up link */
	  bURL=z.getParent();
	  if(bURL!=null)
	  {
	      ans.append("<a href=\"/store/browse?dir=");
	      ans.append(bURL);
	      if(bURL.length()>1)
		  ans.append(File.separatorChar);
	      ans.append("\">[Parent directory]</a><p>\n");
	  }
	  /* subdirs links */
	  String filez[]=f.list();
	  if(filez!=null)
	  {
	      for(int i=0;i<filez.length;i++)
	      {
		  z=new File(f,filez[i]);
		  if(z.isDirectory())
		  {
		      if(del==true)
		      {
			  ans.append("<a href=\"/store/rmdir?dir=");
			  ans.append(URL);
			  ans.append(filez[i]);
			  ans.append(File.separatorChar);
			  ans.append("\">Delete directory</a> ");
		      }

		      ans.append("<a href=\"/store/browse?dir=");
		      ans.append(URL);
		      ans.append(filez[i]);
		      ans.append(File.separatorChar);
		      ans.append("\">[DIR] ");
		      ans.append(filez[i]);
		      ans.append("</a><br>\n");
		  }
	      }
	  }

	  /* objects */
	  bURL=directoryToURL(URL);
	  cachedir d=httpreq.mgr.getDir(mgr.cache_dir+URL,bURL);
	  
	  for(Enumeration en=d.getObjects();en.hasMoreElements();)
	  {
	      cacheobject o;
	      o=(cacheobject)en.nextElement();
	      if(del==true)
	      {
		  ans.append("<a href=\"/store/delete?url=");
		  ans.append(escape(bURL));
		  ans.append(escape(o.getName()));
		  ans.append("\">Del</a> ");
	      }
	      ans.append("<a href=\"");
	      ans.append(bURL);
	      ans.append(o.getName());
	      ans.append("\">");
	      if(o.getName().length()==0)
	      {
		  ans.append("**Welcome**");
	      } else
	        ans.append(o.getName());
	      ans.append(" ");
	      ans.append(o.getType());
	      ans.append(" ");
	      ans.append(o.getSize());
	      ans.append("</a><br>\n");
	  }
      }

      ans.append("<p><form action=\"/store/search\" method=get>\nSearch for: <input type=text name=dir>\nDepth: <input type=text name=depth size=3 maxlength=3 value=2><input type=submit value=\"Search\"></form>\n");
      return ans.toString();
  } else
  if(URL.startsWith("/store/search?"))
  {
      String depth=getparam(URL,"depth");
      if(depth==null) depth="2";
      if(depth.length()==0) depth="2";
      URL=getparam(URL,"dir");
      StringBuffer ans=new StringBuffer(2048);
      if(URL==null)
	  return "Can not search for null string";
      if(URL.length()==0)
	  return "Can not search for empty string";
      ans.append("<title>Smart Cache search</title>\n<h1>Directory search for ");
      ans.append(URL);
      ans.append("</h1>\n");

      search_core(mgr.cache_dir,URL,Integer.valueOf(depth).intValue(),ans);
      ans.append("<p><form action=\"/store/search\" method=get>\nSearch for: <input type=text name=dir value=\"");
      ans.append(URL);
      ans.append("\">Depth: <input type=text name=depth size=3 maxlength=3 value=2><input type=submit value=\"Search\"></form>\n");
      return ans.toString();
  }
  else
  if(URL.startsWith("/store/delete?"))
  {
      URL=getparam(URL,"url");
      if(URL==null) return "Missing URL argument!";
      String parsed[];
      parsed=mgr.parseURL(URL,null);
      /* konverze na local dir */
      String locdir=mgr.getLocalDir(parsed[0],parsed[1],parsed[2],parsed[4]);
      cachedir dir;
      dir=httpreq.mgr.getDir(locdir,parsed[4],parsed[0],parsed[1],parsed[2]);
      cacheobject o=dir.getObject(parsed[3]);
      o.delete();
      return "Object was deleted.";
  } 
  else
  if(URL.startsWith("/store/rmdir?"))
  {
      String bURL;
      URL=getparam(URL,"dir");
      if(URL==null) return "Missing dir argument!";
      if(URL.indexOf("..")>-1) 
	  return "Unsecure directory name.";
      File f=new File(mgr.cache_dir+URL);
      if(URL.length()==0 || !f.isDirectory())
	  return "This is not a valid directory!";
      else
      {
	  bURL=directoryToURL(URL);
	  cachedir d=httpreq.mgr.getDir(mgr.cache_dir+URL,bURL);
	  cacheobject o;
	  for(Enumeration en=d.getObjects();en.hasMoreElements();)
	  {
	      o=(cacheobject)en.nextElement();
	      o.delete();
	  }
	  return "All objects in directory deleted. (Except subdirs)\n";
      }
  } 
  else
  if(URL.startsWith("/ui/refresh?off"))
  {
	      ui.loader_add_reloads=OFF;
	      return "Reload watching has been turned off.";
  }
  else
  if(URL.startsWith("/ui/refresh?on"))
  {
	ui.loader_add_reloads=ON;
        return "Confirmation dialog for Reload watching has been turned off.";
  }
  else
  if(URL.startsWith("/ui/refresh?confirm"))
  {
	ui.loader_add_reloads=ON;
        return "Confirmation dialog for Reload watching has been turned on.";
  } 
  else
  if(URL.startsWith("/ui/add?off"))
  {
	ui.loader_add_missing=OFF;
	return "Adding of missing pages has been turned off.";
  }
  else
  if(URL.startsWith("/ui/add?on"))
  {
	ui.loader_add_missing=OFF;
	return "Auto adding of missing pages has been turned on.";
  }
  else
  if(URL.startsWith("/ui/add?confirm"))
  {
       ui.loader_add_missing=OFF;
       return "Confirmation of adding missing pages has been turned on.";
  } 
  else
  if(URL.equals("/"))
  {
      return "<title>Smart Cache HTML UI</title>\n<h1>Smart Cache UI Home</h1>\n<p>Welcome to Smart Cache HTML User Interface "+VERSION+" running for "+scache.VERSION+".\n<p>You can <a href=\"/store/browse?dir="+File.separatorChar+"\">browse cache</a>, <a href=\"/loader/queue/show\">show current Loader queue</a>.\n"+
      "<h2>Directory search</h2>\n<form action=\"/store/search\" method=get>\nSearch for: <input type=text name=dir>Depth: <input type=text name=depth size=3 maxlength=3 value=2><input type=submit value=\"Search\"></form><p>\n"+
      "<h2>Setup offline support</h2><p>Reload checks: [ <a href=\"/ui/refresh?off\">OFF</a> | <a href=\"/ui/refresh?on\">ON</a> | <a href=\"/ui/refresh?confirm\">CONFIRM</a> ]<br>\nMissing pages: [ <a href=\"/ui/add?off\">OFF</a> | <a href=\"/ui/add?on\">ON</a> | <a href=\"/ui/add?confirm\">CONFIRM</a> ]\n"+
	     "<p><form method=get action=\"/loader/setdepth\">\n"+
	     "Set default depth for load to: <input type=text name=depth maxlength=3 size=3 value=\""+loader_add_depth+"\"><input type=submit value=OK>\n<input type=hidden name=type value=add></form><br>"+
	     "<form method=get action=\"/loader/setdepth\">\n"+
	     "Set default depth for refresh to: <input type=text name=depth maxlength=3 size=3 value=\""+loader_refresh_depth+"\"><input type=submit value=OK>\n<input type=hidden name=type value=refresh></form>";
  }
  
  return "ERROR: Your requested URL "+URL+" was not recogized as UI valid command!<p><a href=\"/\">Go to the homepage</a>.";
}

public final static String getparam(String URL,String arg)
{
    int i1=URL.lastIndexOf(arg+'=');
    if(i1==-1) return null;
    int i2=URL.indexOf('&',i1);
    if(i2==-1) i2=URL.length();
    String part=URL.substring(i1+1+arg.length(),i2);
    return unescape(part);
}

public final static String escape(String URL)
{
    StringBuffer result=new StringBuffer(80);
    char c;
    int l=URL.length();
    for(int i=0;i<l;i++)
    {
	c=URL.charAt(i);
	if(ESCAPE.indexOf(c)>-1)
	{
	    result.append('%');
	    result.append(Integer.toHexString((int)c));
	}
	else
	{
	    result.append(c);
	}
    }
    return result.toString();
}

public final static String unescape(String URL)
{
    StringBuffer result=new StringBuffer(80);
    char c;
    int l=URL.length();
    for(int i=0;i<l;i++)
    {
	c=URL.charAt(i);
	if(c=='%' && i<l-2)
	    try
	    {
		c=(char)Integer.valueOf(URL.substring(i+1,i+3),16).intValue();
		result.append(c);
		i+=2;
	    }
	    catch (NumberFormatException nfe)
	    {
		result.append('%');
		continue;
	    }
	else
	    result.append(c);
    }

    return result.toString();
}

/* konverze localdir na URL */
/* /2/3/www.lakret.cz_2345/ -> http://www.lakret.cz:2345/ */
final static public String directoryToURL(String dir)
{
   String urlpart;
   if(dir.length()==0) return null;
   dir=dir.substring(1);
   /* preskocit 2x file-separator */
   int i;
   i=dir.indexOf(File.separatorChar,0);
   if(i!=-1) {
      dir=dir.substring(i+1);
      i=dir.indexOf(File.separatorChar,0);
      if(i!=-1) dir=dir.substring(i+1);
                else
      return null;
   } else return null;
   i=dir.indexOf(File.separatorChar);
   if(i==-1) return null;
   if(File.separatorChar!='/')
       dir=dir.replace(File.separatorChar,'/');
   urlpart=dir.substring(i);
   String hn=dir.substring(0,i);
   String proto=null;
   String port=null;

  i=hn.indexOf('^');
  if(i>-1)
  {
    proto=hn.substring(i+1);
    hn=hn.substring(0,i);
  }

  i=hn.indexOf('_');
  if(i>-1)
  {
    port=hn.substring(i+1);
    hn=hn.substring(0,i);
  }

  return (proto==null?"http":proto)+"://"+hn+(port==null?"":":"+port)+urlpart;
}

final static Hashtable load_queue()
{
    Hashtable result=new Hashtable(20);
    String line;
    int i;
    StringTokenizer st;
    try
    {
	DataInputStream dis=new DataInputStream(
	                     new BufferedInputStream(
			      new FileInputStream(loader_queue_file)
			     )
			    );
	while( (line=dis.readLine())!=null)
	{
	    st=new StringTokenizer(line);
	    i=st.countTokens();
	    if(i==0) continue;
	    line="";
	    for(int j=0;j<i-1;j++)
	    {
		line+=st.nextToken();
	        line+=" ";
	    }
	    result.put(st.nextToken(),line);
	}
	dis.close();
    }
    catch (IOException grr)
    {}

    return result;
}			    

final static void save_queue(Hashtable data) throws IOException
{
	DataOutputStream dos=new DataOutputStream(
	                     new BufferedOutputStream(
			      new FileOutputStream(loader_queue_file)
			     )
			    );
	String URL;
	for(Enumeration keys=data.keys(); keys.hasMoreElements();)
	{
	    URL=(String)keys.nextElement();
	    
	    dos.write(((String)data.get(URL)).getBytes());
            dos.write(' ');
	    dos.write(URL.getBytes());
            dos.write('\n');
	}
	dos.close();
}			    

public final static void search_core(String dir,String what,int depth,StringBuffer ans)
{
    if(depth<0) return;
    File f=new File(dir);
    File d;
    String filez[]=f.list();
    if(filez==null) return;
    if(filez.length==0) return;
    int rl=mgr.cache_dir.length();
    String bURL;
    for(int z=filez.length-1;z>=0;z--)
    {
	d=new File(f,filez[z]);
	if(!d.isDirectory()) continue;
	bURL=dir+File.separatorChar+filez[z];
	if(filez[z].indexOf(what)>-1)
	{
	    ans.append("<a href=\"/store/browse?dir=");
	    ans.append(dir.substring(rl));
	    ans.append(File.separatorChar);
	    ans.append(filez[z]);
	    ans.append(File.separatorChar);
	    ans.append("\">");
	    ans.append(directoryToURL(bURL.substring(rl)+File.separatorChar));
	    ans.append("</a><br>\n");
	}
	search_core(bURL,what,depth-1,ans);
    }
}

} /* class ui */
